# TODO: Add License

import contextlib
import copy
import json
import logging
import os.path
import pathlib
import pickle
import tempfile
import textwrap
from typing import Any, Dict, List, Iterable, Optional, Union

import pandas as pd
import requests

from .core_objects import (
    AttributionExplanation,
    DatasetInfo,
    ModelInfo,
    MulticlassAttributionExplanation,
)

LOG = logging.getLogger()

SUCCESS_STATUS = 'SUCCESS'
FAILURE_STATUS = 'FAILURE'

FIDDLER_ARGS_KEY = '__fiddler_args__'

ROUTING_HEADER_KEY = 'x-fdlr-fwd'
ADMIN_SERVICE_PORT = 4100


class FiddlerApi:
    """Broker of all connections to the Fiddler API.
    Conventions:
        - Exceptions are raised for FAILURE reponses from the backend.
        - Methods beginning with `list` fetch lists of ids (e.g. all model ids
            for a project) and do not alter any data or state on the backend.
        - Methods beginning with `get` return a more complex object but also
            do not alter data or state on the backend.
        - Methods beginning with `run` invoke model-related execution and
            return the result of that computation. These do not alter state,
            but they can put a heavy load on the computational resources of
            the Fiddler engine, so they should be paralellized with care.
        - Methods beginning with `delete` permanently, irrevocably, and
            globally destroy objects in the backend. Think "rm -rf"
        - Methods beginning with `upload` convert and transmit supported local
            objects to Fiddler-friendly formats loaded into the Fiddler engine.
            Attempting to upload an object with an identifier that is already
            in use will result in an exception being raised, rather than the
            existing object being overwritten. To update an object in the
            Fiddler engine, please call both the `delete` and `upload` methods
            pertaining to the object in question.

    :param url: The base URL of the API to connect to. Usually either
        https://api.fiddler.ai (cloud) or http://localhost:4100 (onebox)
    :param org_id: The name of your organization in the Fiddler platform
    :param auth_token: Token used to authenticate. Your token can be
        created, found, and changed at <FIDDLER ULR>/settings/credentials.
    :param verbose: if True, api calls will be logged verbosely
    """

    def __init__(self, url: str, org_id: str, auth_token: str,
                 verbose=False):

        if url[-1] == '/':
            raise ValueError('url should not end in "/"')

        self.url = url
        self.org_id = org_id
        self.auth_header = {'Authorization': f'Bearer {auth_token}'}
        self.verbose = verbose

        try:
            _ = self.list_projects()
        except requests.exceptions.ConnectionError:
            LOG.warning(
                f'CONNECTION CHECK FAILED: Unable to connect with to '
                f'{self.url}. Are you sure you have the right URL?')
        except RuntimeError as error:
            LOG.warning(f'API CHECK FAILED: Able to connect to {self.url}, '
                        f'but request failed with message "{str(error)}"')

    @staticmethod
    def _get_routing_header(path_base: str) -> Dict[str, str]:
        """Returns the proper header so that a request is routed to the correct
        service."""
        executor_service_bases = (
            'execute',
            'explain',
            'generate',
            'new_project',
        )
        if path_base in executor_service_bases:
            return {ROUTING_HEADER_KEY: 'executor_service'}
        else:
            return {ROUTING_HEADER_KEY: 'data_service'}

    def _call(self, path: List[str], json_payload: Any = None,
              files: Optional[List[pathlib.Path]] = None,
              is_get_request=False):
        """Issues a request to the API and returns the result,
        logigng and handling errors appropriately.


        Raises a RuntimeError if the response is a failure or cannot be parsed.
        Does not handle any ConnectionError exceptions thrown by the `requests`
        library.
        """

        # set up a context manager to open files
        with contextlib.ExitStack() as context_stack:
            endpoint = '/'.join([self.url] + path)
            request_type = 'GET' if is_get_request else 'POST'
            LOG.debug(f'running api call as {request_type} request\n'
                      f'to {endpoint}\n'
                      f'with headers {self.auth_header}\n'
                      f'with payload {json_payload}')
            if is_get_request:
                req = requests.Request('GET', endpoint)
            else:
                # if uploading files, we use a multipart/form-data request and
                # dump the json_payload to be the special "fiddler args"
                # as a json object in the form
                if files is not None:
                    # open all the files into the context manager stack
                    opened_files = {
                        fpath.name: context_stack.enter_context(
                            fpath.open('rb'))
                        for fpath in files
                    }
                    '''
                    NOTE: there are a lot LOT of ways to do this wrong with
                    `requests`

                    Take a look here (and at the thread linked) for more info:
                    https://stackoverflow.com/a/35946962

                    And here: https://stackoverflow.com/a/12385661
                    '''
                    form_data = {
                        **{
                            FIDDLER_ARGS_KEY: (
                                None,                           # filename
                                json.dumps(json_payload),       # data
                                'application/json'              # content_type
                            )
                        },
                        **{
                            fpath.name: (
                                fpath.name,                     # filename
                                opened_files[fpath.name],       # data
                                'application/octet-stream'      # content_type
                            )
                            for fpath in files
                        }
                    }
                    req = requests.Request('POST', endpoint, files=form_data)
                else:
                    req = requests.Request('POST', endpoint, json=json_payload)

            # add necessary headers
            req = req.prepare()
            headers = {**self.auth_header, **self._get_routing_header(path[0])}
            req.headers = {**headers, **req.headers}

            # log the raw request
            raw_request_info = (f'Request:\n'
                                f'  url: {req.url}\n'
                                f'  method: {req.method}\n'
                                f'  headers: {req.headers}')
            LOG.debug(raw_request_info)
            if self.verbose:
                print(raw_request_info)

            # send the request
            res = requests.Session().send(req)

        # catch auth failure
        if res.status_code == 401:
            error_msg = (f'API call to {endpoint} failed with status 401: '
                         f'Authorization Required. '
                         f'Do you have the right org_id and auth_token?')
            LOG.debug(error_msg)
            raise RuntimeError(error_msg)

        # catch non-JSON response (this is rare, the backend should generally
        # return JSON in all cases)
        try:
            response_payload = res.json()
        except json.JSONDecodeError:
            print(res.status_code)
            error_msg = (f'API call to {endpoint} failed.\n'
                         f'Request response: {res.text}\n'
                         f'{raw_request_info}')
            LOG.debug(error_msg)
            raise RuntimeError(error_msg)

        # catch non-auth failures
        try:
            res.raise_for_status()
            assert response_payload['status'] == SUCCESS_STATUS
            # default to None if not returned
            result = response_payload.get('result')
        except Exception:
            response_payload = res.json()
            try:
                failure_exception_message = response_payload['message']
                failure_exception_stacktrace = response_payload['stacktrace']
                error_msg = (
                    f'API call to {endpoint} failed.\n'
                    f'Error message: {failure_exception_message}\n'
                    f'Stacktrace: {failure_exception_stacktrace}\n'
                    f'{raw_request_info}'
                )
            except KeyError:
                error_msg = (f'API call to {endpoint} failed.\n'
                             f'Request response: {res.text}\n'
                             f'{raw_request_info}')
            LOG.debug(error_msg)
            raise RuntimeError(error_msg)

        # log the API call on success (excerpt response on success)
        response_excerpt = textwrap.indent(
            json.dumps(response_payload, indent=2)[:2048], '  ')
        log_msg = (f'API call to {endpoint} succeeded.\n'
                   f'Request response: {response_excerpt}\n'
                   f'{raw_request_info}\n')
        LOG.debug(log_msg)
        return result

    def list_datasets(self) -> List[str]:
        """List the ids of all datasets in the organization.

        :returns: List of strings containing the ids of each dataset.
        """
        path = ['list_datasets', self.org_id]
        return self._call(path, is_get_request=True)

    def list_projects(self) -> List[str]:
        """List the ids of all projects in the organization.

        :returns: List of strings containing the ids of each project.
        """
        path = ['list_projects', self.org_id]
        return self._call(path, is_get_request=True)

    def list_models(self, project_id: str) -> List[str]:
        """List the names of all models in a project.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.

        :returns: List of strings containing the ids of each model in the
            specified project.
        """
        path = ['list_models', self.org_id, project_id]
        return self._call(path, is_get_request=True)

    def get_dataset_info(self, dataset_id: str) -> DatasetInfo:
        """Get DatasetInfo for a dataset.

        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: A fiddler.DatasetInfo object describing the dataset.
        """
        path = ['dataset_schema', self.org_id, dataset_id]
        res = self._call(path, is_get_request=True)
        return DatasetInfo.from_dict(res)

    def get_model_info(self, project_id: str, model_id: str) -> ModelInfo:
        """Get ModelInfo for a model in a certain project.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: A fiddler.ModelInfo object describing the model.
        """
        path = ['model_info', self.org_id, project_id, model_id]
        res = self._call(path, is_get_request=True)
        return ModelInfo.from_dict(res)

    def _query_dataset(self, dataset_id: str, fields: List[str],
                       max_rows: int, split: Optional[str] = None):
        query = dict(fields=fields, limit=max_rows)
        if split is not None:
            query['source'] = f'{split}.csv'
        path = ['dataset_query', self.org_id, dataset_id]
        return self._call(path, query)

    def get_dataset(self, dataset_id: str, max_rows: int = 1_000,
                    splits: Optional[List[str]] = None) \
            -> Dict[str, pd.DataFrame]:
        """Fetches data from a dataset on Fiddler.

        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.
        :param max_rows: Up to this many rows will be fetched from eash split
            of the dataset.
        :param splits: If specified, data will only be fetched for these
            splits. Otherwise, all splits will be fetched.

        :returns: A dictionary of str -> DataFrame that maps the name of
            dataset splits to the data in those splits.
        """
        dataset_info = self.get_dataset_info(dataset_id)
        column_names = dataset_info.get_column_names()
        column_dtypes = dataset_info.get_column_pandas_dtypes()

        def get_df_from_split(split):
            dataset_rows = self._query_dataset(
                dataset_id, fields=column_names, max_rows=max_rows,
                split=split)
            df = pd.DataFrame(dataset_rows, columns=column_names)
            for column_name in df:
                dtype = column_dtypes[column_name]
                df[column_name] = df[column_name].astype(dtype)
            return df

        if splits is None:
            splits = [os.path.splitext(filename)[0]
                      for filename in dataset_info.files]

        return {split: get_df_from_split(split) for split in splits}

    def delete_dataset(self, dataset_id: str):
        """Permanently delete a dataset.

        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine.

        :returns: Server response for deletion action.
        """
        path = ['dataset_delete', self.org_id, dataset_id]
        return self._call(path)

    def delete_model(self, project_id: str, model_id: str):
        """Permanently delete a model.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: Server response for deletion action.
        """
        path = ['delete_model', self.org_id, project_id, model_id]
        return self._call(path)

    def delete_project(self, project_id: str):
        """Permanently delete a project.

        :param project_id: The unique identifier of the project on the Fiddler
            engine.

        :returns: Server response for deletion action.
        """
        path = ['delete_project', self.org_id, project_id]
        return self._call(path)

    def _upload_dataset_csvs(self, dataset_id: str,
                             csv_file_paths: List[pathlib.Path],
                             split_test: Optional[bool] = None,
                             dataset_info: Optional[DatasetInfo] = None):
        """Uploads a CSV file to the Fiddler platform."""
        path = ['dataset_upload', self.org_id]
        payload = dict(dataset_name=dataset_id)
        if split_test is not None:
            payload['split_test'] = split_test
        if dataset_info is not None:
            payload['dataset_info'] = dict(dataset=dataset_info.to_dict())
        return self._call(path, json_payload=payload, files=csv_file_paths)

    def _db_import(self, dataset_id):
        path = ['import_dataset', self.org_id, dataset_id]
        return self._call(path, is_get_request=True)

    def upload_dataset(self,
                       dataset: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                       dataset_id: str,
                       info: Optional[DatasetInfo] = None,
                       split_test=False):
        """Uploads a dataset to the Fiddler engine.

        :param dataset: A DataFrame or dictionary mapping name -> DataFrame
            containing tabular data to be uploaded to the Fiddler engine.
        :param dataset_id: The unique identifier of the dataset on the Fiddler
            engine. Must be a short string without whitespace.
        :param info: A DatasetInfo object specifying all the details of the
            dataset. If not provided, a DatasetInfo will be inferred from the
            dataset and a warning raised.
        :param split_test: If you would like Fiddler to automatically split a
            single-dataframe dataset into a training set and a test set, set
            `split_test=True`. This option has no effect for multi-dataframe
            datasets.

        :returns: The server response for the upload.
        """
        assert ' ' not in dataset_id, \
            'The dataset identifier should not contain whitespace'

        # get a dictionary of str -> pd.DataFrame for all data to upload
        if not isinstance(dataset, dict):
            dataset = dict(data=dataset)

        # infer a dataset_info
        inferred_info = DatasetInfo.from_dataframe(dataset.values(),
                                                   display_name=dataset_id)

        # validate `info` if passed
        if info is not None:
            inferred_columns = inferred_info.get_column_names()
            passed_columns = info.get_column_names()
            if inferred_columns != passed_columns:
                raise RuntimeError(
                    f'Provided data schema has columns {passed_columns}, '
                    f'which does not match the data schema {inferred_columns}')
        # use inferred info with a warning if not `info` is passed
        else:
            LOG.warning(
                f'Heads up! We are inferring the details of your dataset from '
                f'the dataframe(s) provided. Please take a second to check '
                f'our work.'

                f'\n\nIf the following DatasetInfo is an incorrect '
                f'representation of your data, you can construct a '
                f'DatasetInfo with the DatasetInfo.from_dataframe() method '
                f'and modify that object to reflect the correct details of '
                f'your dataset.'

                f'\n\nAfter constructing a corrected DatasetInfo, please '
                f're-upload your dataset with that DatasetInfo object '
                f'explicitly passed via the `info` parameter of '
                f'FiddlerApi.upload_dataset().'

                f'\n\nYou may need to delete the initially uploaded version'
                f'via FiddlerApi.delete_dataset(\'{dataset_id}\').'

                f'\n\nInferred DatasetInfo to check:'
                f'\n{textwrap.indent(repr(inferred_info), "  ")}'

            )
            info = inferred_info

        # determine whether or not the index of this dataset is a meaningful
        # column that should be written into CSV files
        include_index = next(iter(dataset.values())).index.name is not None

        # dump CSVs to named temp file
        with tempfile.TemporaryDirectory() as tmp:
            csv_paths = list()
            for name, df in dataset.items():
                filename = f'{name}.csv'
                path = pathlib.Path(tmp) / filename
                csv_paths.append(path)
                LOG.info(f'[{name}] dataset upload: '
                         f'writing csv to {path}')
                df.to_csv(path, index=include_index)

            # add files to the DatasetInfo on the fly
            new_schema = copy.deepcopy(info)
            new_schema.files = [path.name for path in csv_paths]

            # upload the CSV
            LOG.info(f'[{dataset_id}] dataset upload: uploading csv')
            self._upload_dataset_csvs(dataset_id, csv_paths,
                                      split_test=split_test,
                                      dataset_info=new_schema)

        # trigger dataset upload
        LOG.info(f'[{dataset_id}] dataset upload: '
                 f'triggering import of dataset')
        res = self._db_import(dataset_id)
        return res

    def run_model(self, project_id: str, model_id: str,
                  df: pd.DataFrame) -> pd.DataFrame:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param df: A dataframe contining model inputs as rows.

        :returns: A pandas DataFrame containing the outputs of the model.
        """
        path = ['execute', self.org_id, project_id, model_id]
        payload = dict(data=df.to_dict(orient='records'))
        res = self._call(path, json_payload=payload)
        return pd.DataFrame(res)

    def run_explanation(
            self,
            project_id: str,
            model_id: str,
            df: pd.DataFrame,
            explanations: Union[str, Iterable[str]] = 'shap',
            dataset_id: Optional[str] = None) \
            -> Union[AttributionExplanation,
                     MulticlassAttributionExplanation,
                     List[AttributionExplanation],
                     List[MulticlassAttributionExplanation]]:
        """Executes a model in the Fiddler engine on a DataFrame.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.
        :param df: A dataframe containing model inputs as rows. Only the first
            row will be explained.
        :param explanations: A single string or list of strings specifying
            which explanation algorithms to run.
        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine. Required for most tabular explanations, but
            optional for most text explanations.

        :returns: A single AttributionExplanation if `explanations` was a
            single string, or a list of AttributionExplanation objects if
            a list of explanations was requested.
        """
        """Explains a model's prediction on a single instance"""
        # wrap single explanation name in a list for the API
        if isinstance(explanations, str):
            explanations = (explanations,)
        path = ['explain', self.org_id, project_id, model_id]
        payload = dict(
            data=df.to_dict(orient='records')[0],
            explanations=[dict(explanation=ex) for ex in explanations]
        )
        if dataset_id is not None:
            payload['dataset'] = dataset_id
        res = self._call(path, json_payload=payload)
        explanations_list = res['explanations']
        # TODO: enable more robust check for multiclass explanations
        is_multiclass = 'explanation' not in explanations_list[0]
        deserialize_fn = (MulticlassAttributionExplanation.from_dict
                          if is_multiclass
                          else AttributionExplanation.from_dict)
        ex_objs = [deserialize_fn(explanation_dict)
                   for explanation_dict in explanations_list]
        if len(ex_objs) == 1:
            return ex_objs[0]
        else:
            return ex_objs

    def create_project(self, project_id: str):
        """Create a new project.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine. Must be a short string without whitespace.

        :returns: Server response for creation action.
        """
        path = ['new_project', self.org_id, project_id]
        try:
            res = self._call(path)
        except RuntimeError as e:
            if 'already exists' in str(e):
                print('Project already exists, no change.')
            else:
                raise e
        else:
            return res

    def create_model(self, project_id: str, dataset_id: str,
                     target: str,
                     features: Optional[List[str]] = None,
                     train_splits: Optional[List[str]] = None,
                     model_id: str = 'fiddler_generated_model'):
        """Trigger auto-modeling on a dataset already uploaded to Fiddler.

        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.

        :param dataset_id: The unique identifier of the dataset in the
            Fiddler engine.
        :param target: The column name of the target to be modeled.
        :param features: If specified, a list of column names to use as
            features. If not specified, all non-target columns will be used
            as features.
        :param train_splits: A list of splits to train on. If not specified,
            all splits will be used as training data. Currently only a single
            split can be specified if `train_splits` is not None.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine.

        :returns: Server response for creation action.
        """
        if train_splits is not None:
            if len(train_splits) == 0:
                raise ValueError(
                    '`train_splits` cannot be an empty list. Pass `None` if '
                    'you want to train on the entire dataset.')
            if len(train_splits) > 1:
                raise NotImplementedError(
                    'Sorry, currently only single-split training is '
                    'supported. Please only pass a maximum of one element to '
                    '`train_splits`.')
        source = f'{train_splits[0]}.csv'
        dataset_column_names = self.get_dataset_info(
            dataset_id).get_column_names()

        # raise exception if misspelled target
        if target not in dataset_column_names:
            raise ValueError(f'Target "{target}" not found in the columns of '
                             f'dataset {dataset_id} ({dataset_column_names})')

        # raise if target in features or features not in columns
        if features is not None:
            if target in features:
                raise ValueError(f'Target "{target}" cannot also be in '
                                 f'features.')
            features_not_in_dataset = set(features) - set(dataset_column_names)
            if len(features_not_in_dataset) > 0:
                raise ValueError(f'All features should be in the dataset, but '
                                 f'the following features were not found in '
                                 f'the dataset: {features_not_in_dataset}')

        # use all non-target columns from dataset if no features are specified
        if features is None:
            features = list(dataset_column_names)
            features.remove(target)

        path = ['generate', self.org_id, project_id]
        payload = dict(dataset=dataset_id, model_name=model_id,
                       source=source, inputs=features, targets=[target])
        return self._call(path, json_payload=payload)

    def upload_model_sklearn(
            self, model, info: ModelInfo, project_id: str, model_id: str,
            associated_dataset_ids: Optional[List[str]] = None):
        """Uploads a subclass of sklearn.base.BaseEstimator to the Fiddler
        engine.

        :param model: An instance of a sklearn.base.BaseEstimator object to be
            uploaded. NOTE: custom subclasses are not supported.
        :param info: A ModelInfo object describing the details of the model.
        :param project_id: The unique identifier of the model's project on the
            Fiddler engine.
        :param model_id: The unique identifier of the model in the specified
            project on the Fiddler engine. Must be a short string without
            whitespace.
        :param associated_dataset_ids: The unique identifiers of datasets in
            the Fiddler engine to associate with the model.

        :returns: Server response for upload action.
        """
        assert ' ' not in model_id, \
            'The model identifier should not contain whitespace'

        model_info = copy.deepcopy(info)
        # add associated dataset ids to ModelInfo
        if associated_dataset_ids is not None:
            for dataset_id in associated_dataset_ids:
                assert ' ' not in dataset_id, \
                    'Dataset identifiers should not contain whitespace'
            model_info.misc['datasets'] = associated_dataset_ids

        # add framework info in ModelInfo
        try:
            sklearn_version_number = model.__getstate__()['_sklearn_version']
            info.framework = f'scikit-learn {sklearn_version_number}'
        except KeyError:
            pass

        sklearn_upload_warning = (
            'You are uploading a scikit-learn model using the Fiddler API.'
            '\nIf this model uses any custom (non-sklearn) code, it will not '
            'run properly on the Fiddler Engine.'
            '\nThe Fiddler engine may not be able to detect this in advance.'
        )
        LOG.warning(sklearn_upload_warning)

        with tempfile.TemporaryDirectory() as tmp:
            pickle_path = pathlib.Path(tmp) / 'model.pkl'
            with pickle_path.open('wb') as pickle_file:
                pickle.dump(model, pickle_file)

            # upload the model
            path = ['model_upload', self.org_id]
            payload = dict(
                project=project_id,
                model=model_id,
                model_schema=dict(model=model_info.to_dict()),
                framework=info.framework,
            )
            LOG.info(f'[{model_id}] model upload: uploading pickle')
            return self._call(path, json_payload=payload, files=[pickle_path])
